import math
from scipy.stats import norm,t

def ztest(X:float, mu:float, sd:float,n:float,alpha=0.05,two_tail=False):

  """
  ztest(): Calculates Z test for single sample, two tailed as well as single.

  :param X: Sample mean
  :type X: float
  :param mu: Population mean/ Data mean
  :type mu: float
  :param sd: Standard Deviation
  :type sd: float
  :param n: Sample size
  :type n: float
  :param alpha: P value Limit/Significance value (default:0.05)
  :type alpha: float
  :param two_tail: Enter true for conducting two-tailed test. (default: False)
  :type two_tail: Boolean 
  :return: [Z value, Pvalue, True/False (Test was conclusive or not)]
  :rtype: list
  """
  tail=1

  if(two_tail):
    tail=2
  if(sd==0 or n==0):
    return None
  z=(X-mu)/(sd/(math.sqrt(n)))
  pval=tail*(norm.sf(abs(z)))
  conclusion = (pval<=alpha)

  return [round(z, 3), round(pval, 4), conclusion]

def ztest2(X1:float, X2:float, mudiff:float, sd1:float, sd2:float, n1:float, n2:float,alpha=0.05,two_tail=False):
  """
  ztest2(): Calculates Z test for two sampled data, two tailed as well as single.

  :param X1: Sample mean 1
  :type X: float
  :param X2: Sample mean 2
  :type X: float
  :param mudiff: mean difference
  :type mu: float
  :param sd1: Standard Deviation 1
  :type sd1: float
  :param sd2: Standard Deviation 2
  :type sd2: float
  :param n1: Sample size 1
  :type n1: float
  :param n2: Sample size 2
  :type n2: float
  :param alpha: P value Limit/Significance value (default:0.05)
  :type alpha: float
  :param two_tail: Enter true for conducting two-tailed test. (default: False)
  :type two_tail: Boolean 
  :return: [Z score, P value, True/False (Test was conclusive or not)]
  :rtype: list
  """

  tail=1
  if(two_tail):
    tail=2

  if(n1+n2<=0):
    return None

  pooledSE = math.sqrt(sd1**2/n1 + sd2**2/n2)
  if (pooledSE==0):
    return None
  z = ((X1 - X2) - mudiff)/pooledSE
  pval = tail*norm.sf(abs(z))
  conclusion = (pval<=alpha)
  return [round(z, 3), round(pval, 4), conclusion]

def ztestp(P1:float, P2:float,P:float, mudiff:float, n1:float, n2:float, alpha=0.05,two_tail=False):
  """
  ztestp(): Calculates Z test for two population proportions, two tailed as well as single.

  :param P1: Proportion 1
  :type P1: float
  :param P2: Proportion 2
  :type P2: float
  :param P: Overall Proportion 
  :type P: float
  :param mudiff: mean difference
  :type mudiff: float
  :param n1: Sample size 2
  :type n1: float
  :param n2: Sample size 2
  :type n2: float
  :param alpha: P value Limit/Significance value (default:0.05)
  :type alpha: float
  :param two_tail: Enter true for conducting two-tailed test. (default: False)
  :type two_tail: Boolean 
  :return: [Z value, Pvalue, True/False (Test was conclusive or not)]
  :rtype: list
  """
  tail=1
  if(two_tail):
    tail=2
  if(n1+n2<=0):
    return None
  z = ((P1-P2)-mudiff)/math.sqrt(P*(1-P)*((1/n1)+(1/n2)))
  pval = tail*norm.sf(abs(z))
  conclusion = (pval<=alpha)

  return [round(z, 3), round(pval, 4), conclusion]


def ttest(X:float, mu:float, sd:float, n:float, alpha=0.5,two_tail=False):
  """ttest(): Calculates T test for single sample, two tailed as well as single. This function works for Paired tests as well as the formula is same.

  :param X: Sample mean
  :type X: float
  :param mu: Population mean/ Data mean
  :type mu: float
  :param sd: Standard Deviation
  :type sd: float
  :param n: Sample size
  :type n: float
  :param alpha: P value Limit/Significance value (default:0.05)
  :type alpha: float
  :param two_tail: Enter true for conducting two-tailed test. (default: False)
  :type two_tail: Boolean 
  :return: [T score, P value, True/False (Test was conclusive or not)]
  :rtype: list
  """
  tail=1
  if(two_tail):
    tail=2
  if(sd==0 or n==0):
    return None

  tt=(X-mu)/(sd/(math.sqrt(n)))
  pval=tail*(t.sf(abs(tt),n-1))
  conclusion = (pval<=alpha)
  return [round(tt, 3), round(pval, 4), conclusion]

def ttest2(X1:float, X2:float, mudiff:float, sd1:float, sd2:float, n1:float, n2:float,alpha=0.05,V=False,two_tail=False):
  """ttest2(): Calculates T test for two sampled data (independent), two tailed as well as single.

  :param X1: Sample mean 1
  :type X: float
  :param X2: Sample mean 2
  :type X: float
  :param mudiff: mean difference
  :type mu: float
  :param sd1: Standard Deviation 1
  :type sd1: float
  :param sd2: Standard Deviation 2
  :type sd2: float
  :param n1: Sample size 1
  :type n1: float
  :param n2: Sample size 2
  :type n2: float
  :param alpha: P value Limit/Significance value (default:0.05)
  :type alpha: float
  :param V: Enter True assuming equal variances. (default: False)
  :type V: Boolean 
  :param two_tail: Enter true for conducting two-tailed test. (default: False)
  :type two_tail: Boolean 
  :return: [T score, P value, True/False (Test was conclusive or not)]
  :rtype: list
  """
  tail=1
  if(two_tail):
    tail=2
  if(n1+n2<=0):
    return None
  if(V):
    pooledSE = math.sqrt((((sd1**2)*(n1-1) +(sd2**2)*(n2-1))/(n1+n2-2)))*math.sqrt(1/n1 + 1/n2)
  else:
    pooledSE = math.sqrt(sd1**2/n1 + sd2**2/n2)
  print(pooledSE)


  if (pooledSE==0):
    return None
  tt = ((X1 - X2) - mudiff)/pooledSE
  pval = tail*(t.sf(abs(tt),n1+n2-2))
  conclusion = (pval<=alpha)

  return [round(tt, 3), round(pval, 4), conclusion]







import math 

#Calculation of Future Value given pv,r,m,n
def FV(pv:float, r:float, n:float, m = 1):
  """
  FV(): A function to calculate Future Value.

  :param pv: Present Value
  :type pv: float
  :param r: Rate of Interest/Discount Rate (in decimals eg: 0.05 for 5%)
  :type r: float
  :param n: Number of Years
  :type n: float
  :param m: Frequency of Interest Calculation (eg: 2 for Semi-annually), defaults to 1.
  :type m: float
  :return: float, None for -ve values of m.
  :rtype: float

  --------
  """
  if(m<=0):return None
  return pv*((1+r/m)**(m*n))


#Calculation of Present Value given fv,c,r,m,n
def PV(fv:float, c:float, r:float, n:float, m = 1):
  """
  PV(): A function to calculate Present Value.

  :param fv: Future Value
  :type fv: float
  :param c: Coupon rate (in decimals eg: 0.05 for 5%)
  :type c: float
  :param r: Rate of Interest/Discount Rate (in decimals eg: 0.05 for 5%)
  :type r: float
  :param n: Number of Years
  :type n: float
  :param m: Frequency of Interest Calculation (eg: 2 for Semi-annually), defaults to 1.
  :type m: float
  :return: float, None for -ve values of m.
  :rtype: float

  --------

  """
  if(m<=0):return None
  pv1,pv2=0,0
  pv1 = fv*(c/m)*(1-(1/(1+r/m)**(m*n)))/(r/m)
  pv2=fv/((1+r/m)**(m*n))
  return pv1+pv2

#Calculation of missing variable among fv,pv,r,n
def solve(fv = -1, pv = -1, r = -1, n = -1, m = 1):
  """
  solve(): A function to calculate missing variable among fv,pv,r,n.

  :param fv: Future Value
  :type fv: float
  :param pv: Present Value
  :type pv: float
  :param r: Rate of Interest/Discount Rate (in decimals eg: 0.05 for 5%)
  :type r: float
  :param n: Number of Years
  :type n: float
  :param m: Frequency of Interest Calculation (eg: 2 for Semi-annually), defaults to 1.
  :type m: float
  :return: The function returns the missing variable among fv,pv,r,n and None for -ve values of m. :#Note: ONLY INPUT 3 Parameters to find the fourth one. Parameter 'm' will be set automatically to if left unentered. 
  :rtype: float

  --------
  """


  if(m<=0):return None
  if(fv==-1 and pv!=-1 and r!=-1 and n!=-1):
    fv = FV(pv,r,n,m)
    return fv
  if(pv==-1 and fv!=-1 and r!=-1 and n!=-1):
    pv = PV(fv,0,r,n,m)
    return pv
  if(r==-1 and fv!=-1 and pv!=-1 and n!=-1):
    r = m*(((fv/pv)**(1/m*n)-1))
    return r
  if(n==-1 and fv!=-1 and pv!=-1 and r!=-1):
    n = math.log(fv/pv)/(m*math.log(1+(r/m)))
    return n
  return None



import numpy as np
from scipy.stats import norm
N = norm.cdf

def bsm(S:float, K:float, T:float, r:float, sigma:float,call_value="both"):

    """
    bsm(): Implementation of Black-Scholes formula for non-dividend paying options

    :param S: Current asset price
    :type S: float
    :param K: Strike price of the option
    :type K: float
    :param T: Time until option expiration 
    :type T: float
    :param r: Risk-free interest rate.
    :type r: float
    :param sigma: Annualized volatility of the asset's returns 
    :type sigma: float
    :param call_value: Takes input as "call", "put", "both"
    :type call_value: string
    :return: Price of the options
    :rtype: float. Returns a dictionary when call_value is set to "both"
    """
    d1 = (np.log(S/K) + (r + sigma**2/2)*T) / (sigma*np.sqrt(T))
    d2 = d1 - sigma * np.sqrt(T)
    Value = 0

    if(call_value=="both"):
        Value_call =  S * N(d1) - K * np.exp(-r*T)* N(d2)
        Value_put = K*np.exp(-r*T)*N(-d2) - S*N(-d1)
        return {"put":Value_put, "call":Value_call}

    if(call_value=="call"):
        Value = S * N(d1) - K * np.exp(-r*T)* N(d2)

    elif(call_value=="put"):
        Value = K*np.exp(-r*T)*N(-d2) - S*N(-d1)
    return Value 

    

def bsm_div(S:float, K:float, T:float, r:float, q:float, sigma:float,call_value="both"):

    """
    bsm_div(): Implementation of Black-Scholes formula for Dividend paying options

    :param S: Current asset price
    :type S: float
    :param K: Strike price of the option
    :type K: float
    :param T: Time until option expiration 
    :type T: float
    :param r: Risk-free interest rate.
    :type r: float
    :param q: The dividend rate of the asset.
    :type q: float
    :param sigma: Annualized volatility of the asset's returns. 
    :type sigma: float
    :param call_value: Takes input as "call", "put", "both"
    :type call_value: string
    :return: Price of the options
    :rtype: float. Returns a dictionary when call_value is set to "both"
    """
    d1 = (np.log(S/K) + (r - q + sigma**2/2)*T) / (sigma*np.sqrt(T))
    d2 = d1 - sigma* np.sqrt(T)
    Value = 0

    if(call_value=="both"):
        Value_put = K*np.exp(-r*T)*N(-d2) - S*np.exp(-q*T)*N(-d1)
        Value_call = S*np.exp(-q*T) * N(d1) - K * np.exp(-r*T)* N(d2)
        return {"put":Value_put, "call":Value_call}

    if (call_value=="call"):
        Value = S*np.exp(-q*T) * N(d1) - K * np.exp(-r*T)* N(d2)

    elif(call_value=="put"):
        Value = K*np.exp(-r*T)*N(-d2) - S*np.exp(-q*T)*N(-d1)

    return Value







import math

def IR(spot:list, m=1):
    """
    IR(): A function to calculate Single Effective Interest Rate from an array of spot rates.

    :param spot: An array/List of Spot rates 
    :type spot: list
    :param m: Frequency of Interest Calculation (eg: 2 for Semi-annually), defaults to 1.
    :type m: float
    :return: float, None for -ve values of m.
    :rtype: float
    """
    if(m<=0 or len(spot)==0):
        return None
    return spot[-1]



def FR(spot:list, k:float, m=1):
    """
    FR(): A function to calculate Forward Rate from an array of spot rates.

    :param spot: An array/List of Spot rates 
    :type spot: list
    :param k: Term Period to calculate Forward rate from. (Eg: Nth year forward rate **K** years from now)
    :type k: float
    :param m: Frequency of Interest Calculation (eg: 2 for Semi-annually), defaults to 1.
    :type m: float
    :return: float, None for -ve values of m.
    :rtype: float
    """

    if(m<=0 or k<=0 or len(spot)<=1 or k>len(spot)):
        return None
    
    num = (1+spot[-1]/m)**(len(spot)*m)
    denom = (1+spot[k-1]/m)**(k*m)
    if(denom==0):
        return None

    sum = (num/denom) -1

    return sum
    


from setuptools import find_packages, setup

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='Finstein',
    packages=find_packages(include=['Finstein']),
    version='3.2',
    description='A library for finacial calculations',
    author='Abhiraj Mengade',
    license='MIT',
    install_requires=['scipy','numpy'],
    long_description=long_description,
    long_description_content_type='text/markdown',
    setup_requires=[],
    tests_require=[],
    test_suite='Tests',
)